import re
from itertools import product
from .dolpa_logger import get_logger


LOGGER = get_logger()


def get_dict_value_from_json_path(search_dict, json_path):
    if '.' not in json_path:
        stripped_key = json_path.strip()
        if len(stripped_key) < len(json_path):
            LOGGER.warning(f"Removing extra spaces when using the key {stripped_key}. Correct your Json file...")
        return search_dict[stripped_key]
    path_tokens = json_path.split('.')
    current_value = search_dict
    for token in path_tokens:
        if '[' in token:
            assert ']' in token
            path_token, index = token.split('[')
            index = index[:-1]
            current_value = current_value[path_token][int(index)]
        else:
            current_value = current_value[token]
    return current_value


def get_all_possible_lookouts_from_json_path_with_wild_cards(json_path, lower_limit=0, upper_limit=100):
    expansion_stack = []
    for index, char in enumerate(json_path):
        if char == '[':
            end_index = json_path.index(']', index)
            current_range = json_path[index+1:end_index]
            if ':' in current_range:
                start, end = [int(item) for item in current_range.split(":")]
                expansion_stack.append(list(range(start, end)))
            else:
                if current_range == '*':
                    expansion_stack.append(list(range(lower_limit, upper_limit)))
                else:
                    expansion_stack.append([int(current_range)])
    possible_lookouts = list(product(*expansion_stack))
    return possible_lookouts


def do_dict_interpolation(call: dict, run_config):

    def convert_string_to_value(raw_value):
        if not raw_value.strip().startswith("{{"):
            return raw_value
        variable_name = raw_value.strip()[2:-2].strip()
        return run_config[variable_name]

    if len(call) == 0:
        return {}

    interpolated_dict = {}
    for key, value in call.items():
        if isinstance(value, str):
            interpolated_dict[key] = convert_string_to_value(value)
        elif isinstance(value, list):
            interpolated_list = []
            for list_item in value:
                if isinstance(list_item, str):
                    interpolated_list.append(convert_string_to_value(list_item))
                elif isinstance(list_item, dict):
                    dict_interpolated = do_dict_interpolation(list_item, run_config)
                    interpolated_list.append(dict_interpolated)
                elif isinstance(list_item, list):
                    dict_interpolated = do_dict_interpolation({'placeholder': list_item}, run_config)
                    interpolated_list.append(dict_interpolated['placeholder'])
        elif isinstance(value, dict):
            dict_interpolated = do_dict_interpolation(value, run_config)
            interpolated_dict[key] = dict_interpolated
    return interpolated_dict


def do_string_interpolation(raw_string, run_config):
    if '{{' not in raw_string:
        return raw_string
    interpolated_string = raw_string
    pattern = r'\{\{.*?\}\}'
    matches = re.finditer(pattern, raw_string)
    for match in matches:
        matched_text = match.group()
        actual_value = get_dict_value_from_json_path(run_config, match.group().strip()[2:-2].strip())
        interpolated_string = interpolated_string.replace(matched_text, actual_value, 1)
    return interpolated_string


def interpolate(target, run_config):
    if isinstance(target, str):
        stripped_key = target.strip()
        if len(stripped_key) < len(target):
            LOGGER.warning(f"Removing extra spaces when using the key {stripped_key}. Correct your Json file...")
        return do_string_interpolation(target, run_config)
    if isinstance(target, dict):
        return do_dict_interpolation(target, run_config)



import logging
import sys


class CustomFormatter(logging.Formatter):

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s - %(levelname)s - %(message)s)"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_logger():
    root_logger = logging.getLogger(__name__)
    root_logger.setLevel(logging.DEBUG)

    if not root_logger.hasHandlers():
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(CustomFormatter())
        root_logger.addHandler(handler)
        return root_logger
    return root_logger



class FailedAssertion(Exception):
    pass


class NotAllowedComparison(Exception):
    pass


class InValidCallAttributeError(Exception):
    pass


class InValidCallAttributeModifierError(Exception):
    pass



import os
import json

import requests
from requests.auth import HTTPDigestAuth
from .dolpa_utils import interpolate, get_dict_value_from_json_path
from .dolpa_logger import get_logger
from .comparator import Comparator
from .exceptions import InValidCallAttributeError, InValidCallAttributeModifierError


LOGGER = get_logger()


class EndpointCall:
    allowed_keys = {'identifier', 'description', 'resource', 'method', 'body', 'headers', 'saves', 'assertions'}
    allowed_modifiers = {'replace', 'merge'}

    def __init__(self, call):
        self._validate(call)
        self.identifier = call.get('identifier')
        self.description = call.get('description')
        self.resource = call['resource']
        self.method = call['method']
        self.body = call.get('body')
        self.headers = call.get('headers')
        self.saves = call.get('saves')
        self.assertions = call.get('assertions')
        self._response = None
        self._header_strategy = None
        self._auth_strategy = None

    def _validate(self, call):
        keys_set = {key if ':' not in key else key.split(':')[0] for key in call.keys()}
        modifier_set = {key.split(':')[1] for key in call.keys() if ':' in key}
        if not set(modifier_set).issubset(self.allowed_modifiers):
            invalid_modifiers = set(modifier_set) - (modifier_set.intersection(self.allowed_modifiers))
            raise InValidCallAttributeModifierError(f'{invalid_modifiers} not allowed. Fix your JSON file')
        if not set(keys_set).issubset(self.allowed_keys):
            invalid_attrs = set(keys_set) - (set(keys_set).intersection(self.allowed_keys))
            raise InValidCallAttributeError(f'{invalid_attrs} not allowed. Fix your JSON file.')

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, http_response):
        self._response = http_response


class IntegrationTests:

    call_method_to_req_method_mapping = {
        'GET': requests.get,
        'POST': requests.post,
        'PUT': requests.put,
        'DELETE': requests.delete,
        'HEAD': requests.head,
        'OPTIONS': requests.options,
        'PATCH': requests.patch,
    }

    def __init__(self, runner_dict):
        self.response_store = []
        self.run_config = runner_dict['config']
        self.calls = runner_dict['calls']
        self._env_var_names = ['dolpa_username', 'dolpa_password', 'dolpa_auth_token', 'dolpa_environment']

    def _load_into_config_from_env(self):
        for var_name in self._env_var_names:
            env_value = os.getenv(var_name)
            if env_value:
                self.run_config[var_name] = env_value

    def _combine_global_with_local(self, key, local_map):
        if self.run_config.get(key):
            if local_map and isinstance(local_map, dict):
                return {**self.run_config[key], **local_map}
            else:
                return {**self.run_config[key]}

    def _get_auth(self):
        auth_type = self.run_config.get('auth')
        if not auth_type:
            return None
        if self.run_config['auth'].lower() == 'basic':
            return self.run_config['dolpa_username'], self.run_config['dolpa_password']
        if self.run_config['auth'].lower() == 'digest':
            return HTTPDigestAuth(self.run_config['dolpa_username'], self.run_config['dolpa_password'])
        return None

    def _call_execute(self, call):
        endpoint = self.run_config['base_url'] + call.resource if call.resource.startswith('/') else call.resource
        headers = self._combine_global_with_local('headers', call.headers)
        requests_func = self.call_method_to_req_method_mapping[call.method.upper()]
        response = requests_func(
            endpoint,
            json=interpolate(call.body, self.run_config),
            headers=headers,
            auth=self._get_auth(),
        )
        json_res = response.json()
        for key, val in call.saves.items():
            self.run_config[key] = get_dict_value_from_json_path(json_res, val[2:]) if val.startswith("$.") else val
        for assertion_name, assertion in call.assertions.items():
            Comparator(assertion_name, assertion, self.run_config, json_res).execute()
        return response

    def run_all(self):
        self._load_into_config_from_env()
        for call_dict in self.calls:
            endpoint_call = EndpointCall(call_dict)
            endpoint_call.response = self._call_execute(endpoint_call)

    def run(self, call_identifier):
        self._load_into_config_from_env()
        call = [item for item in self.calls if item["identifier"] == call_identifier][0]
        endpoint_call = EndpointCall(call)
        endpoint_call.response = self._call_execute(endpoint_call)
        return endpoint_call


def run_integration_tests(folder_path):
    with open(folder_path, 'r') as read_file:
        int_test = IntegrationTests(json.load(read_file))
        int_test.run_all()


def run_bulk_integration_tests(root_path):
    contents = sorted(os.listdir(root_path))
    for curr_path in contents:
        inner_path = (root_path + curr_path) if root_path.endswith('/') else (root_path + '/' + curr_path)
        if os.path.isdir(inner_path):
            run_bulk_integration_tests(inner_path)
        else:
            if inner_path.endswith('.json'):
                with open(inner_path, 'r') as read_file:
                    int_test = IntegrationTests(json.load(read_file))
                    int_test.run_all()


def get_integration_test_handler(file_path):
    with open(file_path, 'r') as read_file:
        int_test = IntegrationTests(json.load(read_file))
        return int_test



from .dolpa_utils import interpolate, get_dict_value_from_json_path
from .exceptions import FailedAssertion, NotAllowedComparison
from .dolpa_logger import get_logger

LOGGER = get_logger()


class Comparator:
    def __init__(self, assertion_name, assertion, run_config, response):
        self.assertion_name = assertion_name
        self.assertion = assertion
        self.run_config = run_config
        self.response = response
        self.allowed_comparators = ["==", "<=", ">=", "!=", "<", ">"]

    def execute(self):
        left_side, right_side = None, None
        comparator = None
        for operator in self.allowed_comparators:
            if operator in self.assertion:
                left_side, right_side = (item.strip() for item in self.assertion.split(operator))
                comparator = operator
                break
        if comparator is None:
            raise NotAllowedComparison(f"Found invalid comparator! Valid comparators are f{self.allowed_comparators}")
        if left_side.startswith("$."):
            left_side_interpolated = interpolate(left_side[2:], self.run_config)
        else:
            left_side_interpolated = left_side
        right_side_interpolated = interpolate(right_side, self.run_config)
        if comparator == "==":
            is_pass = get_dict_value_from_json_path(self.response, left_side_interpolated) == right_side_interpolated
        elif comparator == "<=":
            is_pass = get_dict_value_from_json_path(self.response, left_side_interpolated) <= right_side_interpolated
        elif comparator == ">=":
            is_pass = get_dict_value_from_json_path(self.response, left_side_interpolated) >= right_side_interpolated
        elif comparator == "!=":
            is_pass = get_dict_value_from_json_path(self.response, left_side_interpolated) != right_side_interpolated
        elif comparator == "<":
            is_pass = get_dict_value_from_json_path(self.response, left_side_interpolated) < right_side_interpolated
        else:
            is_pass = get_dict_value_from_json_path(self.response, left_side_interpolated) > right_side_interpolated
        if is_pass:
            LOGGER.info(f"{self.assertion_name} which asserts {self.assertion} passed")
        else:
            raise FailedAssertion(f"{self.assertion_name} which asserts {self.assertion} failed")


